@@ -96,59 +96,16 @@ static void register_pam_authenticate_with_login_password_fn(void) {
 	pam_authenticate_with_login_password_fn = APR_RETRIEVE_OPTIONAL_FN(pam_authenticate_with_login_password);
 }
 
-static int hex2char(int c) {
-	if (c >= '0' && c <= '9')
-		return c - '0';
-	if (c >= 'a' && c <= 'z')
-		return c - 'a' + 10;
-	if (c >= 'A' && c <= 'Z')
-		return c - 'A' + 10;
-	return -1;
-}
-
 static char * intercept_form_submit_process_keyval(apr_pool_t * pool, const char * name,
 	const char * key, int key_len, const char * val, int val_len) {
 	if (val_len == 0)
 		return NULL;
-	int i;
-	for (i = 0; i < key_len; i++, name++) {
-		if (*name == '\0')
-			return NULL;
-		int c = key[i];
-		if (c == '+')
-			c = ' ';
-		else if (c == '%') {
-			if (i > key_len - 3)
-				return NULL;
-			int m = hex2char(key[++i]);
-			int n = hex2char(key[++i]);
-			if (m < 0 || n < 0)
-				return NULL;
-			c = (m << 4) + n;
-		}
-		if (c != *name)
-			return NULL;
-	}
-	if (*name != '\0')
+	char * x_key = apr_pstrndup(pool, key, key_len);
+	ap_unescape_urlencoded(x_key);
+	if (strcmp(name, x_key))
 		return NULL;
-	char * ret = apr_palloc(pool, val_len + 1);
-	char * p = ret;
-	for (i = 0; i < val_len; i++, p++) {
-		if (val[i] == '+')
-			*p = ' ';
-		else if (val[i] == '%') {
-			if (i > val_len - 3)
-				return NULL;
-			int m = hex2char(val[++i]);
-			int n = hex2char(val[++i]);
-			if (m < 0 || n < 0)
-				return NULL;
-			*p = (m << 4) + n;
-		} else {
-			*p = val[i];
-		}
-	}
-	*p = '\0';
+	char * ret = apr_pstrndup(pool, val, val_len);
+	ap_unescape_urlencoded(ret);
 	return ret;
 }
 
@@ -181,6 +138,10 @@ static void intercept_form_redact_password(ap_filter_t * f, ifs_config * config)
 	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, SHOW_MODULE "will redact password (value of %s) in the POST data", config->password_name);
 	ifs_filter_ctx_t * ctx = f->ctx;
 	apr_bucket * b = ctx->password_fragment_start_bucket;
+	if (! b) {
+		ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, SHOW_MODULE "the input got processed from the cached_brigade (possibly by ap_send_error_response), no redacting possible");
+		return;
+	}
 	int fragment_start_bucket_offset = ctx->password_fragment_start_bucket_offset;
 	if (fragment_start_bucket_offset) {
 		apr_bucket_split(b, fragment_start_bucket_offset);
@@ -188,7 +149,7 @@ static void intercept_form_redact_password(ap_filter_t * f, ifs_config * config)
 	}
 	char * new_password_data = apr_pstrcat(r->pool, config->password_name, "=", _REDACTED_STRING, NULL);
 	int new_password_data_length = strlen(new_password_data);
-	apr_bucket * new_b = apr_bucket_immortal_create(new_password_data, new_password_data_length, f->c->bucket_alloc);
+	apr_bucket * new_b = apr_bucket_pool_create(new_password_data, new_password_data_length, r->pool, f->c->bucket_alloc);
 	APR_BUCKET_INSERT_BEFORE(b, new_b);
 
 	int password_remove_length = 0;
@@ -201,6 +162,9 @@ static void intercept_form_redact_password(ap_filter_t * f, ifs_config * config)
 		if (b == APR_BRIGADE_SENTINEL(ctx->cached_brigade)) {
 			break;
 		}
+		if (APR_BUCKET_IS_EOS(b)) {
+			break;
+		}
 		if (APR_BUCKET_IS_METADATA(b))
 			continue;
 		const char * buffer;
@@ -304,6 +268,7 @@ static apr_status_t intercept_form_submit_filter(ap_filter_t * f, apr_bucket_bri
 		APR_BRIGADE_CONCAT(bb, ctx->cached_brigade);
 		apr_brigade_cleanup(ctx->cached_brigade);
 		ctx->cached_brigade = NULL;
+		ctx->password_fragment_start_bucket = NULL;
 		return ctx->cached_ret;
 	}
 	return ap_get_brigade(f->next, bb, mode, block, readbytes);
@@ -316,7 +281,6 @@ static apr_status_t intercept_form_submit_filter_prefetch(request_rec * r, ifs_c
 	ifs_filter_ctx_t * ctx = f->ctx;
 	if (! ctx) {
 		f->ctx = ctx = apr_pcalloc(r->pool, sizeof(ifs_filter_ctx_t));
-		ctx->cached_brigade = apr_brigade_create(f->c->pool, f->c->bucket_alloc);
 	}
 
 	char * login_value = NULL;
@@ -329,16 +293,15 @@ static apr_status_t intercept_form_submit_filter_prefetch(request_rec * r, ifs_c
 
 	authn_status out_status = AUTH_GENERAL_ERROR;
 
-	apr_bucket_brigade * bb = apr_brigade_create(f->c->pool, f->c->bucket_alloc);
+	ctx->cached_brigade = apr_brigade_create(f->c->pool, f->c->bucket_alloc);
 	int fetch_more = 1;
 	while (fetch_more) {
-		ctx->cached_ret = ap_get_brigade(f->next, bb, AP_MODE_READBYTES, APR_BLOCK_READ, HUGE_STRING_LEN);
+		ctx->cached_ret = ap_get_brigade(f->next, ctx->cached_brigade, AP_MODE_READBYTES, APR_BLOCK_READ, HUGE_STRING_LEN);
 		if (ctx->cached_ret != APR_SUCCESS)
 			break;
 
-		apr_bucket * b = APR_BRIGADE_FIRST(bb);
-		APR_BRIGADE_CONCAT(ctx->cached_brigade, bb);
-		for (; b != APR_BRIGADE_SENTINEL(ctx->cached_brigade); b = APR_BUCKET_NEXT(b)) {
+		for (apr_bucket * b = APR_BRIGADE_FIRST(ctx->cached_brigade);
+			b != APR_BRIGADE_SENTINEL(ctx->cached_brigade); b = APR_BUCKET_NEXT(b)) {
 			if (! fetch_more)
 				break;
 			if (APR_BUCKET_IS_EOS(b)) {
@@ -364,15 +327,17 @@ static apr_status_t intercept_form_submit_filter_prefetch(request_rec * r, ifs_c
 			while ((nbytes > 0) && (e = memchr(p, '&', nbytes))) {
 				if (fragment) {
 					int new_length = fragment_length + (e - p);
-					fragment = realloc(fragment, new_length);
-					memcpy(fragment + fragment_length, p, e - p);
+					char * new_fragment = apr_palloc(r->pool, new_length);
+					memcpy(new_fragment, fragment, fragment_length);
+					memcpy(new_fragment + fragment_length, p, e - p);
 					if (intercept_form_submit_process_buffer(f, config, &login_value, &password_value,
-						fragment, new_length, fragment_start_bucket, fragment_start_bucket_offset, &out_status)) {
+						new_fragment, new_length, fragment_start_bucket, fragment_start_bucket_offset, &out_status)) {
 						fetch_more = 0;
 						break;
 					}
-					free(fragment);
 					fragment = NULL;
+					fragment_length = 0;
+					fragment_start_bucket = NULL;
 				} else {
 					if (intercept_form_submit_process_buffer(f, config, &login_value, &password_value,
 						p, e - p, b, (p - buffer), &out_status)) {
@@ -388,8 +353,10 @@ static apr_status_t intercept_form_submit_filter_prefetch(request_rec * r, ifs_c
 			if (nbytes > 0) {
 				if (fragment) {
 					int new_length = fragment_length + nbytes;
-					fragment = realloc(fragment, new_length);
-					memcpy(fragment + fragment_length, p, nbytes);
+					char * new_fragment = apr_palloc(r->pool, new_length);
+					memcpy(new_fragment, fragment, fragment_length);
+					memcpy(new_fragment + fragment_length, p, nbytes);
+					fragment = new_fragment;
 					fragment_length = new_length;
 				} else if (APR_BUCKET_NEXT(b) && APR_BUCKET_IS_EOS(APR_BUCKET_NEXT(b))) {
 					/* shortcut if this is the last bucket, slurp the rest */
@@ -397,7 +364,7 @@ static apr_status_t intercept_form_submit_filter_prefetch(request_rec * r, ifs_c
 						p, nbytes, b, (p - buffer), &out_status);
 					fetch_more = 0;
 				} else {
-					fragment = malloc(nbytes);
+					fragment = apr_palloc(r->pool, nbytes);
 					memcpy(fragment, p, nbytes);
 					fragment_length = nbytes;
 					fragment_start_bucket = b;
@@ -406,8 +373,6 @@ static apr_status_t intercept_form_submit_filter_prefetch(request_rec * r, ifs_c
 			}
 		}
 	}
-	if (fragment)
-		free(fragment);
 	return out_status == AUTH_GRANTED ? OK : DECLINED;
 }
 
